/*
 * A filesystem directory resource.
 *
 * Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <pthread-l4.h>

#include <e2access/format.h>
#include <e2access/image.h>
#include <fsclient/client_fcntl.h>
#include <ipc/server.h>
#include <ipc/util_ipc.h>
#include "fs_directory_listing.h"

/* Client and server interfaces. */

#include "directory_listing_client.h"
#include "directory_listing_private_server.h"



/* Callback function for ext2fs_dir_iterate. */

static int read_proc(struct ext2_dir_entry *dir_entry, int offset,
                     int blocksize, char *buf, void *priv_data)
{
  Fs_directory_listing *self = reinterpret_cast<Fs_directory_listing *>(priv_data);

  (void) offset;
  (void) blocksize;
  (void) buf;

  /* Invoke the buffer population method, potentially replying to the original
     message and waiting for further requests. */

  return self->record_entry(dir_entry);
}

/* File type conversion. */

static int convert_file_type(int type)
{
  switch (type)
  {
    case EXT2_FT_REG_FILE:  return DT_REG;
    case EXT2_FT_DIR:       return DT_DIR;
    case EXT2_FT_CHRDEV:    return DT_CHR;
    case EXT2_FT_BLKDEV:    return DT_BLK;
    case EXT2_FT_FIFO:      return DT_FIFO;
    case EXT2_FT_SOCK:      return DT_SOCK;
    case EXT2_FT_SYMLINK:   return DT_LNK;
    default:                return DT_UNKNOWN;
  }
}



/* Thread function. */

static void *listing_thread(void *data)
{
  Fs_directory_listing *obj = reinterpret_cast<Fs_directory_listing *>(data);

  obj->mainloop();
  return 0;
}



/* Administration methods. */

void Fs_directory_listing::close()
{
  /* Notify the thread of the termination. */

  client_DirectoryListing listing(_thread_ep);

  listing.terminate();

  /* Wait for the thread to terminate. */

  pthread_join(_thread, NULL);
}

long Fs_directory_listing::start()
{
  pthread_create(&_thread, NULL, listing_thread, this);

  long err = ipc_server_new_gate_for_thread(&_thread_ep, pthread_l4_cap(_thread), l4_umword_t(this));
  if (err)
    return err;

  return L4_EOK;
}

/* Operation methods. */

size_t Fs_directory_listing::read_into(size_t start, char *buffer, size_t length)
{
  size_t size;

  _buffer = buffer;
  _buffer_size = length;

  client_DirectoryListing listing(_thread_ep);

  if (listing.read(start, &size))
    return 0;
  else
    return size;
}

void Fs_directory_listing::mainloop()
{
  /* Wait for parameters. */

  _wait();

  while (!_terminating)
  {
    /* Always start at the beginning of the listing. */

    _start_pos = 0;
    _index = 0;

    /* Iterate over directory nodes, calling the indicated callback function,
       populating the buffer, returning data to the caller periodically. */

    if (ext2fs_dir_iterate(_fs, _dir, 0, 0, read_proc, this))
      break;

    /* Test termination conditions before assuming that iteration completed
       normally. */

    if (_terminating)
      break;

    /* Restart iteration if the next available data is after the desired
       data. */

    if (_seek_pos < _start_pos)
      continue;

    /* Add a null entry. */

    int need_page = !add_null_entry();

    /* Reply with the final page of data. */

    reply();

    /* Test for a termination request. */

    if (_terminating)
      break;

    /* Expect a request seeking to the end of the results, iterating again
       otherwise. */

    if (_seek_pos != _start_pos)
      continue;

    /* Reply with no data at the end of the results, yielding control and
       obtaining more parameters. */

    if (need_page)
      reply();
    else
      _yield(0);
  }
}

/* Helper function to ensure alignment in generated entries. */

static int pad_align(int value)
{
  return value + (sizeof(unsigned int) - (value % sizeof(unsigned int)));
}

/* Obtain the details of the directory entry for output. */

void Fs_directory_listing::init_entry(struct entry *entry, struct ext2_dir_entry *dir_entry)
{
  int namelen = ext2fs_dirent_name_len(dir_entry);
  int reclen = DIRENT_CORE_SIZE + namelen;

  entry->dir_entry = dir_entry;
  entry->namelen = namelen;

  /* Align the size of the entry to avoid problems on architectures which
     require aligned accesses and where the compiler needs to assume an aligned
     structure. At least one padding byte is added for convenience. */

  entry->reclen = pad_align(reclen);
  entry->size = entry->reclen;

  if (_flags & OPENDIR_STAT)
  {
    /* Stat information added to each entry should have an aligned size. */

    entry->size += sizeof(struct stat);
    image_stat_inode(_fs, entry->dir_entry->inode, &entry->statbuf);
  }
  else
    memset(&entry->statbuf, 0, sizeof(struct stat));
}

/* Populate the buffer associated with the message set on the object, returning
   codes to direct the node iteration being performed by libext2fs. */

int Fs_directory_listing::record_entry(struct ext2_dir_entry *dir_entry)
{
  struct ext2_inode inode;
  struct entry entry;
  size_t start_portion, end_portion;

  init_entry(&entry, dir_entry);

  /* Obtain the inode details for metadata. */

  if (ext2fs_read_inode(_fs, dir_entry->inode, &inode))
    return DIRENT_ABORT;

  /* Iterate until the sought position is reached. */

  if (_seek_pos >= _start_pos + entry.size)
  {
    _start_pos += entry.size;
    return 0;
  }

  /* If the requested start position is within the current entry, discard any
     data from the entry prior to the sought position. */

  if (_seek_pos > _start_pos)
  {
    start_portion = _seek_pos - _start_pos;
    end_portion = entry.size - start_portion;

    copy_entry_end(&entry, start_portion);

    /* Set the next location in the buffer. */

    _index += end_portion;

    /* Set the buffer start position and continue iterating. */

    _start_pos = _seek_pos;

    return 0;
  }

  /* Test for enough space in the buffer. */

  if (_index + entry.size <= _buffer_size)
  {
    /* Copy the entry into the next location in the buffer. */

    copy_entry(&entry);

    /* Set the next location in the buffer and continue iterating. */

    _index += entry.size;

    return 0;
  }

  /* Space in the buffer is exceeded by the entry. */

  start_portion = _buffer_size - _index;
  end_portion = entry.size - start_portion;

  /* Copy as much as fits of the entry into the buffer. */

  copy_entry_start(&entry, start_portion);

  _index += start_portion;

  /* Return the details of the completed entries. */

  reply();

  /* Test for a termination request. */

  if (_terminating)
    return DIRENT_ABORT;

  /* Test the requested start position. If it is before the current entry,
     terminate and restart iteration. */

  else if (_seek_pos < _start_pos - start_portion)
    return DIRENT_ABORT;

  /* If the requested start position is after the current entry, move past
     this entry and continue iterating. */

  else if (_seek_pos >= _start_pos + end_portion)
  {
    _start_pos += end_portion;
    return 0;
  }

  /* The requested start position is within the current entry. Discard any data
     from the entry prior to the sought position. */

  start_portion += _seek_pos - _start_pos;
  end_portion = entry.size - start_portion;
  _start_pos = _seek_pos;

  /* Copy the remainder of the last entry. */

  copy_entry_end(&entry, start_portion);

  /* Set the next location in the buffer. */

  _index += end_portion;

  /* Continue iterating. */

  return 0;
}

/* Add a null entry to the buffer. */

int Fs_directory_listing::add_null_entry()
{
  if (_index + DIRENT_CORE_SIZE > _buffer_size)
    return 0;

  memset(_buffer + _index, 0, DIRENT_CORE_SIZE);
  return 1;
}

/* Copy a complete directory entry into the buffer. */

void Fs_directory_listing::copy_entry(struct entry *entry)
{
  _copy_entry(_buffer + _index, entry);
}

/* Copy a complete directory entry. */

void Fs_directory_listing::_copy_entry(char *out, struct entry *entry)
{
  struct dirent *out_entry = (struct dirent *) out;

  out_entry->d_ino = entry->dir_entry->inode;
  out_entry->d_off = 0;

  /* The record length concerns only the entry details. Any following metadata
     will not be included. */

  out_entry->d_reclen = entry->reclen;
  out_entry->d_type = convert_file_type(ext2fs_dirent_file_type(entry->dir_entry));
  memcpy(out_entry->d_name, entry->dir_entry->name, entry->namelen);

  /* Pad the memory after the name to the alignment boundary. */

  memset(out_entry->d_name + entry->namelen, 0, entry->reclen - entry->namelen);

  /* Write metadata after the entry details if requested. */

  if (_flags & OPENDIR_STAT)
    memcpy(out + entry->reclen, &entry->statbuf, sizeof(struct stat));
}

/* Copy the start of a directory entry into the buffer. */

void Fs_directory_listing::copy_entry_start(struct entry *entry, int size)
{
  char temp[entry->size];

  _copy_entry(temp, entry);
  memcpy(_buffer + _index, temp, size);
}

/* Copy the end of a directory entry into the buffer. */

void Fs_directory_listing::copy_entry_end(struct entry *entry, int offset)
{
  char temp[entry->size];

  _copy_entry(temp, entry);
  memcpy(_buffer + _index, temp + offset, entry->size - offset);
}



/* Return control from copying entries into the buffer. */

void Fs_directory_listing::reply()
{
  /* Pad an incomplete buffer. This is useful when the actual length of the
     written data cannot be determined by the user of the data. */

  memset(_buffer + _index, 0, _buffer_size - _index);

  /* Update the start position of the next page. */

  _start_pos += _buffer_size;

  /* Deliver the end position to the caller, waiting for the new start position
     before continuing. */

  _yield(_buffer_size);

  /* Start any new page from the beginning. */

  _index = 0;
}



/* Wait for parameters. */

void Fs_directory_listing::_wait()
{
  l4_umword_t label = 0;
  ipc_message_t msg;

  /* Provide item expectations for the message. */

  ipc_message_expect(&msg, DirectoryListingPrivate_expected_items);

  while ((label & ~3UL) != l4_umword_t(this))
    ipc_message_wait(&msg, &label);

  /* Invoke the appropriate method. */

  handle_DirectoryListingPrivate(&msg, this);
}

/* Yield control and wait to regain control again. */

void Fs_directory_listing::_yield(size_t read)
{
  /* Complete the read operation, reply and wait.
     NOTE: A special "reply" label is used. */

  complete_DirectoryListingPrivate_read(L4_INVALID_CAP | L4_SYSF_REPLY, read);
  _wait();
}

long Fs_directory_listing::read(size_t position)
{
  _seek_pos = position;
  return L4_EOK;
}

long Fs_directory_listing::terminate()
{
  _terminating = true;
  return L4_EOK;
}
